local testFiltering = false

local messageTypes = {
	["LFPROF"] = LFMonitor.lftxt,
	["LFW"] = LFMonitor.lfwtxt,
	["LFM"] = LFMonitor.lfmtxt,
	["LFG"] = LFMonitor.lfgtxt,
	["WTB"] = LFMonitor.wtbtxt,
	["WTS"] = LFMonitor.wtstxt
}

-- The following code is adapted from r61 of Misspelled, downloaded from
-- http://wow.curse.com/downloads/wow-addons/details/misspelled/download/526785.aspx

if string.utf8len == nil then
  require("libs\\UTF8\\utf8data")
  require("libs\\UTF8\\utf8")
end

--Local reference to Global functions for speed
local string_find = string.find
local table_insert = table.insert
local string_len = string.len
local string_lower =  string.lower
local string_sub = string.sub
local math_min = math.min

--split a string, at delimiter into a table
-- Compatibility: Lua-5.1
local split_t
local split_fpat
local split_last_end
local split_s, split_e, split_cap
local utf8 = false

function wordSimilarity(source, target)

	if utf8 == true then
	  string_len = string_utf8len
      string_lower = string_utf8lower
	  string_sub = string_utf8sub
	end

	local m = string_len(source)
	local n = string_len(target)
	local maxDist = math_min(m,n,5)

	if math.abs(m-n) >= maxDist then
		return 0
	end

	local rows = {{0}}
	m=m+1
	n=n+1
	for x=1, m do
		rows[x] = {0}
		for y=1,n do
			rows[x][y] = 0
		end
	end

	--tprint(rows)

	for i=1,m do
		rows[i][1] = i
	end

	for j=1,n do
		rows[1][j] = j
	end

	--Ignore if the strings that only differ by case
	source = string_lower(source)
	target = string_lower(target)

	--tprint(rows)
	local subsitutionCost

	for i=2, m do
		for j=2,n do
			if string_sub(source, i, i) == string_sub(target, j, j) then
				subsitutionCost = 0
			else
				subsitutionCost = 1
			end

			rows[i][j] = math_min(	rows[i-1][j] + 1, 					--Insertion
									rows[i][j-1] + 1, 					--Deletion
									rows[i-1][j-1] + subsitutionCost)	--Substitution
		end
	end

	local dist = rows[m][n]

	if dist >= maxDist then
		return 0
	else
		return 1 - dist/maxDist
	end
end

-- end of adapted code

-- *****************
-- * Filter Engine *
-- *****************

--replace links with the link text
local function unlink(msg)

	local pre, linktext, post = string.match(msg, "(.*)|c%x+|H[^|]+|h%[([^|]+)%]|h|r(.*)")
	local unlinked

	pre = pre or ""
	post = post or ""

	if (linktext) then

		unlinked = string.lower(unlink(pre) .. " " .. linktext .. " " .. unlink(post))

	end

	return unlinked or string.lower(msg)

end

--function to add spaces around {star},{square}, etc so they are parsed correctly
local function formatspecial(msg)

	local out = string.gsub(msg, "{", " {")

	out = string.gsub(out, "}", "} ")

	return out

end

local function doSearch(profTxt, msgTxt, typeTxt, sender, guid, omsg)

	local threshold = 0.75

	local lmsg = unlink(msgTxt)
	local pattern, message, pattern2 = nil, nil, nil
	local matched = false
	local profmatch = 0
	local typematch = 0

	if (cfflag == "CUSTOM") then
		for _, ttxt in ipairs(typeTxt) do
			if (string.match(lmsg, ttxt)) then
				return true
			end
		end
		return false
	end

	pattern = ".*(www%.).*"
	message = string.match(lmsg, pattern)
	matchedvalue = nil

	if (message ~= nil) then

		--found probable spam, so ignore
		matchquality = 0

	end

	if (profTxt[1] ~= nil) then --nop
		profmatch = 1
	else
		profmatch = fuzzyMessageMatch(lmsg, profTxt)
	end
	typematch = fuzzyMessageMatch(lmsg, typeTxt)
	return (profmatch >= threshold) and (typematch >= threshold)
end

-- ***********************
-- * Debug Functions *
-- ***********************

local function debug(msg)

    DEFAULT_CHAT_FRAME:AddMessage(msg, 1, 0, 0)

end

-- **********************************************
-- * Fuzzy matching & classifier implementation *
-- **********************************************

local function StringHash(text)
  local counter = 1
  local len = string.len(text)
  for i = 1, len, 3 do
    counter = math.fmod(counter*8161, 4294967279) +  -- 2^32 - 17: Prime!
  	  (string.byte(text,i)*16776193) +
  	  ((string.byte(text,i+1) or (len-i+256))*8372226) +
  	  ((string.byte(text,i+2) or (len-i+256))*3932164)
  end
  return math.fmod(counter, 4294967291) -- 2^32 - 5: Prime (and different from the prime in the loop)
end

local function cacheLookup(table, item, class, fallback, disableCaching)
  if disableCaching then
    return fallback()
  end
  local n
  if not table.item then
    table.item = {}
	for n=1,table.max do
	  table.item[n] = false
	end
  end
  if not table.score[class] then
    table.score[class] = {}
	for n=1,table.max do
	  table.score[class][n] = false
	end
  end
  local hashKey = math.fmod(StringHash(item), table.max)
  if table.item[hashKey] == item and table.score[class][hashKey] then
    return table.score[class][hashKey]
  else
    local score = fallback()
	table.item[hashKey] = item
	table.score[class][hashKey] = score
	return score
  end
end

local function fuzzyWordMatch(currentWord, candidates)
  currentWord = string_lower(currentWord)
  local exactMatch = candidates[currentWord]
  if exactMatch then
    return math.max(exactMatch,0)
  end
  local length = string_len(currentWord)
  if length < 3 or length > 20 then
    -- don't match 1- or 2-letter abbrs inexactly, and don't
	  -- waste CPU cycles on words that are too long to match anything
    return 0
  end
  local strongestNegMatch = 0
  local strongestPosMatch = 0
  for candidateWord, strength in pairs(candidates.fuzzy) do
    local candidateLength = string_len(candidateWord)
	matchStrength = wordSimilarity(currentWord,candidateWord) * strength
	if matchStrength > strongestNegMatch then
		strongestPosMatch = matchStrength
	elseif matchStrength < strongestNegMatch then
		strongestNegMatch = matchStrength
	end
  end
  if strongestPosMatch > -strongestNegMatch then
	return strongestPosMatch
  else
    return 0
  end
end

local function fuzzyMessageMatch(words, candidates, className)
  local matchStrength = 0
  local word
  for _, word in pairs(words) do
    matchStrength = matchStrength + fuzzyWordMatch(word, candidates)
  end
  return matchStrength
end

local function classifyMessage(message, classifier)
  message = formatspecial(unlink(message))
  local words = {}
  for word in string.gmatch(message, LFMonitor.WORD_PATTERN) do -- split into words
    tinsert(words,word)
  end
  if testFiltering then
    debug("Message: "..message)
  end
  local threshold = 0.5
  local score
  local scores = {}
  local mayBelongTo = {["unknown"] = true}
  local highestScore = 0.8
  local scoreCap = 1.5
  for class, candidates in pairs(classifier) do
    score = fuzzyMessageMatch(words, candidates, class, disableCaching)
	if score > highestScore then
	  highestScore = score
	end
	scores[class] = score
  end
  local cutoff = math_min(highestScore, scoreCap) * threshold
  for class, score in pairs(scores) do
    if testFiltering and score > 0 then
	  debug("Class: "..class.." Score: "..score)
	end
    if score >= cutoff then
	  mayBelongTo[class] = true
	  mayBelongTo["unknown"] = false
	end
  end
  return mayBelongTo
end

local function cachingClassifyMessage(message, classifier, metaClassName)
  return cacheLookup(LFMonitorCache, message, metaClassName, function() return classifyMessage(message, classifier, disableWordCaching) end)
end

local function isInDesirableClass(message, classifier, desired, metaClassName)
  local classes = cachingClassifyMessage(message, classifier, metaClassName, false)
  for class, matches in pairs(classes) do
    if matches and desired[class] == 1 then
	  return true
	end
  end
  return false
end

local function isDesiredProfession(message)
  return isInDesirableClass(message, LFMonitor.classifierProflist, LFMonitorDB.professions, "prof")
end

local function isDesiredRaid(message)
  return LFMonitorDB.options.overrideraid == 1 or isInDesirableClass(message, LFMonitor.classifierRaidlist, LFMonitorDB.raids, "raid")
end

local function isDesiredCharacterClass(message)
  return isInDesirableClass(message, LFMonitor.classifierClasslist, LFMonitorDB.classes, "class")
end

local function isDesiredRole(message)
  return isDesiredRole(message, LFMonitor.classifierRolelist, LFMonitorDB.roles, "role")
end

-- ************************
-- * Variable Declaration *
-- ************************

local alertframe, cfflag, ddmenu, clmenu, pname, pflag, itemlink = nil, nil, nil, nil, nil, nil
local mouse, buttons = false, false
local inactive, combat = false, false
local tradechannel = -1
local _G = _G
local eventframe = nil
local events = {}
local matchedval = nil

-- ***************************
-- * Miscellaneous functions *
-- ***************************
local function sendWhisper(priority, prefix, msg, type, lang, dest)

	local ch, msgchunk, segment, m
	local space = 0
	local msgs = {}

	msgchunk = msg

	if (string.len(msg) > 255) then

		while (string.len(msgchunk) > 255) do

			segment = string.sub(msgchunk, 1, 255)

			for ch = 255, 1, -1 do

				if (string.sub(segment, ch, ch) == " ") then

					space = ch
					break

				end

			end

			if (space == 0) then

				space = 255

			end

			table.insert(msgs, string.sub(segment, 1, space - 1))
			msgchunk = string.sub(msgchunk, space + 1)

		end

	end

	table.insert(msgs, msgchunk)

	for _, m in ipairs(msgs) do

		ChatThrottleLib:SendChatMessage(priority, prefix, m, type, lang, dest)

	end

end

local function resetEvent()

	if (LFMMAIN:GetScript("OnUpdate") == nil) then

		LFMMAIN:SetScript("OnUpdate", function(self, elapsed) LFMonitor:AlertOnUpdate(self, elapsed) end)

	end

end

function LFMonitor:PlayAlert(id)

	local sound = LFMonitor.sounds[id]

	if (LFMonitorDB.options.soundon == 1) then

		if (sound.type == "game") then

			PlaySound(sound.file)

		else

			PlaySoundFile("Interface\\AddOns\\LFMonitor\\Sounds\\" .. sound.file)

		end

	end

end

local function GetAutoReply()

	local reply

	if (matchedvalue) then

		reply = LFMonitorDB.responses[matchedvalue]

	end

	return reply

end

function LFMonitor:ChatCommand(input)

	local command = string.lower(input or "")

	if (command == "show") then

		LFMonitorDB.options.disabled = 0
		SetDisable(false)

	elseif (command == "hide") then

		LFMonitorDB.options.disabled = 1
		SetDisable(true)

	else

		InterfaceOptionsFrame_OpenToCategory(_G.LFMonitorOptionsMain)

	end

end

local function GetLocalTime()

	local format = LFMonitor.timestamps[LFMonitorDB.options.timestamp].format
	local localtime = date(format)

	return localtime

end

local function IsInstance()

	local inInstance, instanceType = IsInInstance()
	local ininstance = false

	if (inInstance ~= nil) then

		if (instanceType ~= "none") then

  			ininstance = true

		end

	end

  	return ininstance

end

local function alertHide()

	alertframe:Hide()

end

local function respond(self, event, ...)

	if (events[event]) then

		events[event](event, ...)

	end

end

function LFMonitor:lfmframeOnUpdate(self, elapsed)

	local flash = LFMEndFlash
	local flashtimer = LFMonitor.frame.flashtimer + elapsed

	if (alertframe:AtBottom()) then

		if (flash:IsShown()) then

			flash:Hide()

		end

		LFMonitor.frame:SetScript("OnUpdate", nil)

	elseif (flashtimer < CHAT_BUTTON_FLASH_TIME ) then

		LFMonitor.frame.flashtimer = flashtimer

	else

		LFMonitor.frame.flashtimer = 0

		if (flash:IsShown()) then

			flash:Hide()

		else

			flash:Show()

		end

	end

end

function LFMonitor:GetChannelNumber(channel)

	local channels = {GetChannelList()}
	local channelnumber = -1

	for num = 1, #channels, 2 do

		if (channels[num + 1] == channel) then

			channelnumber = num

		end

	end

	return channelnumber

end

local function eventframeInit(init)

	if (LFMonitor.testmode == true) then

		eventframe:RegisterEvent("CHAT_MSG_WHISPER_INFORM")

	else

		eventframe:RegisterEvent("CHAT_MSG_CHANNEL")
		eventframe:RegisterEvent("CHAT_MSG_SAY")
		eventframe:RegisterEvent("CHAT_MSG_YELL")

	end

	if (not inactive) then

		eventframe:RegisterEvent("PLAYER_UPDATE_RESTING")
		eventframe:RegisterEvent("PLAYER_ENTERING_WORLD")
		eventframe:RegisterEvent("PLAYER_REGEN_DISABLED")
		eventframe:RegisterEvent("PLAYER_REGEN_ENABLED")
		eventframe:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		eventframe:RegisterEvent("ZONE_CHANGED")
		eventframe:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE")

	end

	inactive = false
	eventframe:SetScript("OnEvent", respond)

	if (init) then

		tradechannel = LFMonitor:GetChannelNumber("Trade")
		LFMonitor:DisableCheck()

	end

	alertframe:Clear()

end

local function SetDisable(disable)

	if (disable == true) then

		if (alertframe:IsVisible()) then

			LFMonitor:OnDisable()
			inactive = true

		end

	else

		if (not alertframe:IsVisible()) then

			UIFrameFadeIn(alertframe, 1)
			resetEvent()
			eventframeInit()
			inactive = false

		end

	end

end

function LFMonitor:DisableCheck()

	local disable = false

	tradechannel = LFMonitor:GetChannelNumber("Trade")

	if (LFMonitorDB.options.hinst == 1) then

		if (IsInstance() == 1) then

			disable = true

		end

	end

	if (LFMonitorDB.options.rested == 1) then

		if (not IsResting()) then

			disable = true

		end

	end

	if (LFMonitorDB.options.trade == 1) then

		if (tradechannel == -1) then

			disable = true

		end

	end

	SetDisable(disable)

end

function LFMonitor:IsInBounds()

	local scale = UIParent:GetEffectiveScale()
	local x, y = GetCursorPosition()
	local xmin = LFMonitorDB.alert.position.x
	local xmax = xmin + LFMonitorDB.alert.position.width
	local ymax = LFMonitorDB.alert.position.y
	local ymin = ymax - LFMonitorDB.alert.position.height
	local inbounds = nil

	x = x / scale
	y = y / scale

	if ((x >= xmin) and (x <= xmax) and (y >= ymin) and (y <= ymax)) then

		inbounds = true

	end

	return inbounds

end

-- *************************
-- * Alert Frame Functions *
-- *************************

function LFMonitor:AlertShowButtons()

	UIFrameFadeIn(ScrollAlertFrameUpBtn, 0.5, 0, 1)
	UIFrameFadeIn(ScrollAlertFrameDownBtn, 0.5, 0, 1)
	UIFrameFadeIn(ScrollAlertFrameEndBtn, 0.5, 0, 1)

end

function LFMonitor:AlertHideButtons()

	ScrollAlertFrameUpBtn:Hide()
	ScrollAlertFrameDownBtn:Hide()
	ScrollAlertFrameEndBtn:Hide()

end

local function checkHistory(msg, sender)

--ignore duplicate messages sent within 20 seconds of each other
--to try and cut down on channel spam

	local i, v
	local valid = 1

	if (#LFMonitor.history > 0) then

		for i,v in pairs(LFMonitor.history) do

			if ((v.msender == sender) and (v.mmsg == msg)) then

				if (GetTime() - v.mtime < 20) then

					valid = nil

				end

			end

		end

	end

	if (valid) then

		if (LFMonitor.hindex == 10) then

			LFMonitor.hindex = 1

		end

		table.insert(LFMonitor.history, LFMonitor.hindex, {msender = sender, mmsg = msg, mtime = GetTime()})
		LFMonitor.hindex = LFMonitor.hindex + 1

	end

	return valid

end

local function addAlert(msg, sender, guid)
	local hlink
	local plink = "[" .. sender .. "]:|h|r "
	local tlink = "|c" .. LFMonitorDB.alert.font.timestamp .. GetLocalTime() .. "|r "
	local mcol = "|c" .. LFMonitorDB.alert.font.text
	local pcol

	if guid == -99999 then

		return

	end

	if (LFMonitorDB.options.useclasscolours == 1) then

		pcol = LFMonitor:GetClassColour(guid)

	else

		pcol = LFMonitorDB.alert.font.player

	end

	hlink = "|c" .. pcol .. "|Hplayer:" .. sender .. ":LFMONITOR:" .. cfflag .. "|h"

	if (checkHistory(msg, sender)) then

		LFMonitor:PlayAlert(LFMonitorDB.options.sound)

		if (alertframe:GetAlpha() ~= 1) then

			LFMonitor:AlertFadeIn(0.1)

		end

		alertframe.TimeSinceLastUpdate = 0
		alertframe:AddMessage(tlink .. hlink .. plink .. mcol .. LFMonitor:FormatMessage(msg) .. "|r", 1, 1, 1, 1, 60.0)

	end

end

function LFMonitor:SetLocked(locked)

	alertframe.isLocked = locked
	LFMonitorDB.options.locked = locked

end

function LFMonitor:AlertFadeIn(ftime)

	if (not mouse) then

		if (alertframe:GetAlpha() ~= 1) then

			UIFrameFadeIn(alertframe, ftime, 0.3, 1)
			resetEvent()

		end

	end

end

function LFMonitor:AlertFadeOut(ftime)

	if (alertframe:GetAlpha() == 1) then

		UIFrameFadeOut(alertframe, ftime, 1, 0.3)

	end

end

function LFMonitor:SetMouse(bValue)

	mouse = bValue

end

function LFMonitor:SetAlertFont()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.font.text)
	local fontfile = LFMonitor.fonts[LFMonitorDB.alert.font.index].file

	alertframe:SetFont(fontfile, LFMonitorDB.alert.font.height)

end

function LFMonitor:SetAlertBackdropColour()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.backdrop.background)

	alertframe:SetBackdropColor(r, g, b, a)

end

function LFMonitor:SetAlertBorderColour()

	local r, g, b, a = LFMonitor:HexToRGB(LFMonitorDB.alert.backdrop.border)

	alertframe:SetBackdropBorderColor(r, g, b, a)

end

-- *******************
-- * Event Functions *
-- *******************

function LFMonitor:OnInitialise()

	local x, y
	local slashcmd, slashidx = "LFMONITOR_CMD", 1

	SLASH_LFMONITOR_CMD1 = "/" .. LFMonitor.CHAT_COMMANDS[1]
	SLASH_LFMONITOR_CMD2 = "/" .. LFMonitor.CHAT_COMMANDS[2]
	SlashCmdList["LFMONITOR_CMD"] = function(input) LFMonitor:ChatCommand(input) end

	LFMDBSVARS = LFMonitorDB

	if (IsAddOnLoaded("TradeForwarder")) then

		LFMonitor.tfw = true

	else

		LFMonitor.tfw = nil

	end

	--remove old saved data that is no longer required
	if (LFMonitorDB.craftlist) then

		LFMonitorDB.craftlist = nil

	end

end

function LFMonitor:OnDisable()

	local r, g, b, a = LFMMAIN:GetBackdropColor()

	if (LFMonitor.testmode == true) then

		eventframe:UnregisterEvent("CHAT_MSG_WHISPER_INFORM")

	else

		eventframe:UnregisterEvent("CHAT_MSG_CHANNEL")
		eventframe:UnregisterEvent("CHAT_MSG_SAY")
		eventframe:UnregisterEvent("CHAT_MSG_YELL")

	end

	LFMonitor.frame:SetScript("OnUpdate", nil)
	LFMMAIN:SetScript("OnUpdate", nil)

	UIFrameFadeOut(alertframe, 1, a, 0)
	alertframe.fadeInfo.finishedFunc = alertHide
	inactive = true

end

function events.PLAYER_REGEN_DISABLED()

	--player entering combat
	if (LFMonitorDB.options.combat == 1) then

		if (alertframe:IsVisible()) then

			dis = true
			combat = true
			UIFrameFadeOut(alertframe, 1)
			alertframe.fadeInfo.finishedFunc = alertHide

		end

	end

end

function events.PLAYER_REGEN_ENABLED()

	--player leaving combat
	if (combat) then

		UIFrameFadeIn(alertframe, 1)
		resetEvent()
		combat = false

	end

end

function events.PLAYER_UPDATE_RESTING(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.ZONE_CHANGED_NEW_AREA(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.ZONE_CHANGED(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function events.PLAYER_ENTERING_WORLD(event, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

	if (LFMonitorDB.raids.eoe) then

		LFMonitorDB.raids = {}

	end

end

function LFMonitor:AlertOnUpdate(self, elapsed)

	self.TimeSinceLastUpdate = self.TimeSinceLastUpdate + elapsed

	if (self.TimeSinceLastUpdate < LFMonitor.LFM_Update_Int) then

		return

	end

	if (alertframe:IsVisible()) then

		if (alertframe:GetAlpha() == 1) then

			if (mouse == false) then

				self.TimeSinceLastUpdate = 0
				LFMonitor:AlertHideButtons()
		    	LFMonitor:AlertFadeOut(2)
		    	LFMonitor:AlertFadeOut(2)
				LFMMAIN:SetScript("OnUpdate", nil)
				mouse = false

		  	end

		end

	end

end

function LFMonitor:OnHyperlinkShow(self, link, text, button)

	SetItemRef(link, text, button)

end

local function tableEmpty(table)

	for _, _ in pairs(table) do
		return false
	end
	return true

end

local function ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, _, counter, guid, ...)

	local message = nil
	local pattern, lmsg = nil, nil
	local omsg, classtxt = nil, nil
	local lfmclasses = {"dk", "dr", "hu", "ma", "pa", "pr", "ro", "sh", "wa", "wr"}
	local profs = {"jc", "enc", "lw", "ins", "tlr", "bs", "alc", "eng", "misc"}
	local misc = {"lfg", "wtb", "lf", "wts", "lfw"}
	local matched, lfmgen = false, true
	local channelmatch = false
	local tmp = {}
	local i, channels = {}
	local tfwidx, tfwname, tfwmsg, tfwcolour
	local lfmsender, original

	lfmsender = sender

	if (LFMonitorDB.channels.general == 1) then

		if ((zoneid == 1) or (event == "CHAT_MSG_SAY") or (event == "CHAT_MSG_YELL")) then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.trade == 1) then

		if (zoneid == 2) then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.guild == 1) then

		if (event == "CHAT_MSG_GUILD") then

			channelmatch = true

		end

	end

	if (LFMonitorDB.channels.lfg == 1) then

		if (zoneid == 26) then

			channelmatch = true

		end

	end

	if (channelmatch == false) then

		channels = {GetChannelList()}

		for i = 1, #channels, 2 do

			if (channels[i] > 4) then

				if (LFMonitorDB.channels["cc" .. channels[i + 1]] == 1) then

					if (channelnumber == channels[i]) then

						channelmatch = true

					end

				end
			end

		end

	end

	if (LFMMAIN:IsVisible()) then

		if ((channelmatch == true) or (LFMonitor.testmode == true)) then

			original = chatmessage
			omsg = chatmessage

			if LFMonitor.tfw then

				tfwidx, tfwname, tfwmsg, tfwcolor = strmatch(chatmessage, LFMonitor.TFW_PATTERN)

				if tfwmsg then

					omsg = tfwmsg
					lfmsender = tfwname

				end

			end

--			if (not tableEmpty(LFMonitor.proflist)) then

			local messageType = cachingClassifyMessage(omsg, messageTypes, "type", false)

			if messageType["WTB"] and LFMonitorDB.monitoring.wtb == 1 then
				cfflag = "WTB"
				matched = true
			elseif messageType["WTS"] and LFMonitorDB.monitoring.wts == 1 then
				cfflag = "WTS"
				matched = true
			end
			if (not matched) and (messageType["LFPROF"]) then
				cfflag = "PROF"
				matched = LFMonitorDB.monitoring.lf or isDesiredProfession(omsg)
			end
			if (not matched) and (messageType["LFW"]) then
				cfflag = "PROF"
				matched = LFMonitorDB.monitoring.lfw or isDesiredProfession(omsg)
			end
			if (not matched) and messageType["LFG"] and LFMonitorDB.monitoring.lfg == 1 then
				cfflag = "LFM"
				matched = isDesiredRaid(omsg)
			end
			if (not matched) and messageType["LFM"] and LFMonitorDB.monitoring.lfm == 1 then
				matched = isDesiredRaid(omsg) and
				          (isDesiredCharacterClass(omsg) or isDesiredRole(omsg))
			end
			if (not matched) then

				if ((LFMonitorDB.options.basic or "") ~= "") then
					omsg = formatspecial(unlink(omsg))

					cfflag = "CUSTOM"
					tmp = {}

					if (string.find(LFMonitorDB.options.basicesc, " ")) then

						tmp = {strsplit(" ", string.lower(LFMonitorDB.options.basicesc))}

					else

						table.insert(tmp, string.lower(LFMonitorDB.options.basicesc))

					end

					matched = doSearch(LFMonitor.nop, omsg, tmp, lfmsender, guid, original)

				end

			end

			if (not matched) then

				if ((LFMonitorDB.options.adv or "") ~= "") then
					omsg = formatspecial(unlink(omsg))

					cfflag = "CUSTOM"
					tmp = {}
					table.insert(tmp, LFMonitorDB.options.adv)
					matched = doSearch(LFMonitor.nop, omsg, tmp, lfmsender, guid, original)

				end

			end
			if matched then
				addAlert(original, lfmsender, guid)
			end

		end

	end

end

function events.CHAT_MSG_CHANNEL(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_WHISPER_INFORM(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_YELL(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_SAY(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	ChatParser(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

end

function events.CHAT_MSG_CHANNEL_NOTICE(event, chatmessage, sender, language, channelstring, target, flags, zoneid, channelnumber, channelname, counter, ...)

	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

end

function LFMonitor:OnEnable()

	LFMonitor:CreateOptions()

	if (LFMonitorDB.alert.position.x == -1) then

		x = (UIParent:GetWidth() - LFMonitorDB.alert.position.width) / 2
		y = (UIParent:GetHeight() - LFMonitorDB.alert.position.height) / 2
		LFMonitorDB.alert.position.x = x
		LFMonitorDB.alert.position.y = y

	end

	alertframe = LFMonitor:CreateAlertFrame()

	eventframe = CreateFrame("Frame", "LFMonitorEventFrame")
	eventframeInit(true)
	LFMonitor:BuildTables()
	tradechannel = LFMonitor:GetChannelNumber("Trade")
	LFMonitor:DisableCheck()

	if (IsAddOnLoaded("TradeForwarder")) then

		LFMonitor.tfw = 1

	else

		LFMonitor.tfw = nil

	end

end

-- ******************
-- * Dropdown Menus *
-- ******************

local function LFMonitor_DDMenuItem_OnClick(self, button, down)

	local opt = self.value
	local reply, rtmp = nil, nil
	local cl = nil

	PlaySound("igMainMenuOption")

	if (opt == 1) then

		--autoreply
		reply = GetAutoReply()

		if (reply == nil or reply == "") then

			--nothing to autoreply with
			StaticPopup_Show("LFMONITOR_NOAUTOREPLY")

		else

			rtmp = string.sub(reply, 1, 255)

			sendWhisper("NORMAL", "LFMONITOR", reply, "WHISPER", nil, pname)

		end

	elseif(opt == 2) then

		--whisper
		ChatFrame1EditBox:SetText(LFMonitor.WHISPER_COMMAND .. " " .. pname .. " ")

	elseif(opt == 3) then

		--ignore
		AddIgnore(pname)

	elseif(opt == 4) then

		--cancel
		--do nothing

	end

end

local function LFMonitor_DD_Initialise()

	local info

	info = {}
	info.text = pname
	info.value = 0
	info.owner = ddmenu
	info.isTitle = true
	info.func = nil
	info.notCheckable = 1
	UIDropDownMenu_AddButton(info)

	info = {}
	info.text = LFMonitor.AUTOREPLY
	info.value = 1
	info.owner = ddmenu
	info.notCheckable = 1
	info.isTitle = nil

	if (pflag == "LFG" or pflag == "WTB") then

		info.disabled = true

	else

		info.disabled = nil

	end

	info.tooltipTitle = LFMonitor.AUTOREPLY
	info.tooltipText = string.format(LFMonitor.SEND_AUTO, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.WHISPER
	info.value = 2
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.notCheckable = 1
	info.tooltipTitle = _G.WHISPER
	info.tooltipText = string.format(LFMonitor.WHISPER, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.IGNORE
	info.value = 3
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.tooltipTitle = _G.IGNORE
	info.notCheckable = 1
	info.tooltipText = string.format(LFMonitor.IGNORE_LIST, pname)
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

	info = {}
	info.text = _G.CANCEL
	info.value = 4
	info.owner = ddmenu
	info.isTitle = nil
	info.disabled = nil
	info.notCheckable = 1
	info.tooltipTitle = _G.CANCEL
	info.tooltipText = LFMonitor.CLOSE_MENU
	info.func = LFMonitor_DDMenuItem_OnClick
	UIDropDownMenu_AddButton(info, 1)

end

local function LFMonitor_ShowDDMenu(link)

	local htype, hname, hmod, flag = strsplit(":", link)

	pname = hname
	pflag = flag

	ddmenu = _G.LFMonitorDDMenu
	ddmenu.initialize = LFMonitor_DD_Initialise
	ddmenu.displayMode = "MENU"
	ddmenu.name = "LFMonitor"
	PlaySound("igMainMenuOpen")
	ToggleDropDownMenu(1, nil, ddmenu, "cursor")

end

-- *********
-- * Hooks *
-- *********

--Hook SetItemRef so LFMonitor can handle its own custom links
local ORIG_SetItemRef = _G.SetItemRef

local function LFMonitor_SetItemRef (link, text, button, ...)

	local htype, hname, hmod, flag = strsplit(":", link)
	local runorig = nil

	if (IsModifiedClick("CHATLINK")) then

		runorig = true

	elseif (button == "RightButton") then

		if (hmod == "LFMONITOR") then

			LFMonitor_ShowDDMenu(link)

		else

			runorig = true

		end

	else

		runorig = true

	end

	if (runorig) then

		ORIG_SetItemRef(link, text, button, ...)

	end

end

_G.SetItemRef = LFMonitor_SetItemRef

--Hook ChatEdit_GetActiveWindow so LFMonitor can capture links in the options frame
local orig_ChatEdit_GetActiveWindow = ChatEdit_GetActiveWindow

function ChatEdit_GetActiveWindow()

    return LFMonitor.EditFocus or orig_ChatEdit_GetActiveWindow()

end

local function benchmark(operation, message)
	debugprofilestart()
	for i=1,10000 do
		operation()
	end
	local run = debugprofilestop()
	local memory = collectgarbage("count")
	debugprofilestart()
	collectgarbage()
	local gc = debugprofilestop()
	debug(string.format("%s: %d ms (%d run, %d gc), %d kB",message,gc+run,run,gc,memory))
end

function LFMonitor:benchmarkFiltering(messages, cacheSize)
	testFiltering = false
	local testClassifiers = {
		["type"] = messageTypes,
		["role"] = LFMonitor.classifierRolelist,
		["class"] = LFMonitor.classifierClasslist,
		["raid"] = LFMonitor.classifierRaidlist,
		["profession"] = LFMonitor.classifierProflist
	}
	for name, classifier in pairs(testClassifiers) do
		debug("Testing classifier: "..name)
		LFMonitorCache = {score = {}, max = cacheSize}
		cachingClassifyMessage("", testClassifiers["type"]) -- force initialization before the benchmark
		collectgarbage()
		benchmark( function ()
				classifyMessage(testMessages[math.random(#testMessages)], classifier)
			end, "No caching")
		benchmark( function ()
				cachingClassifyMessage(testMessages[math.random(#testMessages)], classifier, namee)
			end, "Message caching")
	end
end
